import asyncio
import typing as t

from coges.coge import Coge
from loguru import logger


class MachineMeta(t.TypedDict):
    tick: t.Any
    active_coges: list[str]
    results: dict[str, t.Any]


TickFn = t.AsyncGenerator[t.Any, t.Any]
State = t.Union[t.Dict[str, t.Any], MachineMeta]
Machine = t.Callable[[], t.Coroutine[t.Any, t.Any, None]]


def identity(x: t.Any) -> t.Any:
    return x


def to_result(name_task: tuple[str, asyncio.Task]) -> tuple[str, bool]:
    return (name_task[0], name_task[1].result())


def by_result(name_result: tuple[str, bool]) -> bool:
    return name_result[1]


def to_name(name_result: tuple[str, bool]) -> str:
    return name_result[0]


async def choose_active_coges(
    coges: list[Coge], state: State, di_resolver
) -> list[Coge]:
    tasks: dict[str, asyncio.Task] = {}
    logger.debug("starting coges predicates")
    async with asyncio.TaskGroup() as tg:
        for _coge in coges:
            logger.debug(f"starting '{_coge.name}' predicate")
            tasks[_coge.name] = tg.create_task(
                di_resolver(_coge.get_predicate())(**state)
            )
    logger.debug("filtering coges by predicate results")
    result = list(
        filter(
            lambda coge: coge.name
            in map(
                to_name,
                filter(
                    by_result,
                    map(to_result, ((name, task) for name, task in tasks.items())),
                ),
            ),
            coges,
        )
    )
    logger.debug(f"finished filtering coges by predicate results: {result}")
    return result


def validate_coges(coges: list[Coge]) -> None:
    empty_predicates = list(filter(lambda coge: coge.get_predicate() is None, coges))
    empty_actions = list(filter(lambda coge: coge.get_action() is None, coges))

    if len(empty_predicates) > 0:
        raise ValueError(f"Coges with empty predicates: {empty_predicates}")

    if len(empty_actions) > 0:
        raise ValueError(f"Coges with empty actions: {empty_actions}")


def create_machine(
    coges: list[Coge], tick_fn: TickFn, di_resolver=identity, initial_state: State = {}
) -> Machine:
    validate_coges(coges)

    state: State = dict(tick=None, results=dict(), active_coges=list(), **initial_state)
    logger.debug(f"starting machine with coges: {coges}")

    async def __machine():
        async for tick in tick_fn:
            logger.debug(f"new tick: '{tick}'")
            state.update({"tick": tick})
            logger.debug(f"choosing active coges for tick '{tick}'")
            active_coges = await choose_active_coges(coges, state, di_resolver)
            logger.debug(f"active coges for tick '{tick}': {active_coges}")
            results: dict[str, t.Any] = {}
            logger.debug(f"running active coges for tick '{tick}'")
            async with asyncio.TaskGroup() as tg:
                for _coge in active_coges:
                    logger.debug(f"running '{_coge.name}'")
                    results[_coge.name] = tg.create_task(
                        di_resolver(_coge.get_action())(**state)
                    )
                logger.debug(f"waiting for results of tick '{tick}'")

            logger.debug(f"finished waiting for results of tick '{tick}'")
            state.update(
                dict(
                    results=dict(
                        **{name: task.result() for name, task in results.items()}
                    )
                )
            )
            state.update({"active_coges": [coge.name for coge in active_coges]})
            logger.debug(f"tick {tick} results in: {state}")

    return __machine



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['coges']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.7.2,<0.8.0']

setup_kwargs = {
    'name': 'coges',
    'version': '0.9.0',
    'description': 'Library for implementing non-linear behaviour and stream processing',
    'long_description': '* Coges\nLibrary for implementing non-linear behaviour and stream processing\n\n** Features\n- Predicate -> Action based flow control\n- Can be used as multi-state and single-state finite machine\n- Built-in dependency injector with lazy resolving and dependency tree\n- Functional approach\n- Concurrent predicate and action running\n- State checks are connected to your stream\n- Zero dependencies\n\n\n* Changelog (0.9.0)\n- tick now should be asyncgenerator\n',
    'author': 'Jellyfish.tech',
    'author_email': 'manager@jellyfish.tech',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import typing as t


State = t.Dict[str, t.Any]
PredicateFn = t.Callable[..., t.Coroutine[t.Any, t.Any, bool]]
ActionFn = t.Callable[..., t.Coroutine[t.Any, t.Any, t.Any]]


class CogeError(Exception):
    pass


class Coge:
    name: str
    __predicate: PredicateFn | None = None
    __action: ActionFn | None = None

    def __init__(self, name: str):
        self.name = name

    def predicate(self, predicate_fn: PredicateFn) -> None:
        self.__predicate = predicate_fn

    def action(self, action_fn: ActionFn) -> None:
        self.__action = action_fn

    def get_predicate(self) -> PredicateFn | None:
        return self.__predicate

    def get_action(self) -> ActionFn | None:
        return self.__action

    def __repr__(self):
        return f"<Coge: {self.name} predicate={self.get_predicate()} action={self.get_action()}>"


def create_coge(name: str) -> Coge:
    return Coge(name)



import logging
import inspect
from functools import lru_cache, partial

logger = logging.getLogger(__name__)


def __async_partial(fn, *args, **kwargs):
    async def inner(*fargs, **fkwargs):
        return await fn(*args, *fargs, **kwargs, **fkwargs)

    return inner


# TODO: maybe value of a dependency itself might be inner class with __call__, __index__, __getattr__ etc methods implemented. That will allow
# TODO: dependencies to be truly lazy and waking up only on usage, not when they are present in parameters
class Dependency:
    def __init__(self, resolver):
        self.name = ""
        self.__resolver = resolver
        self.__factory = (None for _ in range(1))
        self.__instance = None
        self.__value = None

    def __wake_up(self):
        self.__instance = self.__resolver(self.factory)()
        self.__value = next(self.__instance)

    # FIXME: that's a mess of properties, handle in next iteration
    @property
    def factory(self):
        return self.__factory

    @factory.setter
    def factory(self, factory):
        self.__factory = factory

    @property
    def instance(self):
        return self.__instance

    @instance.setter
    def instance(self, value):
        self.__instance = value

    @instance.deleter
    def instance(self):
        if self.__instance is None:
            raise RuntimeError(f"{self.name} was never used")

        try:
            next(self.__instance)
        except StopIteration:
            del self.__instance
        else:
            raise RuntimeError(
                f"{self.name} instance is not exhausted. Factory should contain only one yield statement"
            )

    @property
    def value(self):
        self.__wake_up() if self.__instance is None else None
        return self.__value

    def __repr__(self):
        return "<Dependency: {}>".format(self.name)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name

    def __del__(self):
        del self.instance


def create_dependency_injector():  # noqa: C901 sue me
    dependencies: set[Dependency] = set()

    def resolve(fn):
        fn_parameters = inspect.signature(fn).parameters
        available_dependencies = filter(lambda d: d.name in fn_parameters, dependencies)

        partial_kind = __async_partial if inspect.iscoroutinefunction(fn) else partial

        return partial_kind(fn, **{d.name: d.value for d in available_dependencies})

    def add_dependency(name_or_factory):
        new_dependency = Dependency(resolver=resolve)

        if inspect.isgeneratorfunction(name_or_factory):
            new_dependency.name = name_or_factory.__name__
            new_dependency.factory = name_or_factory

            dependencies.add(new_dependency)

            return name_or_factory

        if isinstance(name_or_factory, str):

            def decorator(fn):
                new_dependency.name = name_or_factory
                new_dependency.factory = fn

                dependencies.add(new_dependency)

                return fn

            return decorator

    return add_dependency, resolve
